#!/usr/bin/env python
# coding: utf-8

import sys
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.axes as maxes

import cartopy as cart

import cartopy.crs as ccrs
import cartopy.feature as cfeature
from mpl_toolkits.axes_grid1 import make_axes_locatable

import csv
import math

import xarray as xr


#------------------------------------------------------
# This notebook was created by Marie Doutriaux-Boucher: marie.doutriauxboucher@eumetsat.int
# April 2024
#------------------------------------------------------
 
def mmm_print(var,texte):
    print(texte+' no filter:',np.min(var),np.max(var),np.mean(var),np.median(var),np.std(var))
    return

proj_dic = {'Robin':ccrs.Robinson()}

#===========================================
def create_map(plot_file, lat, lon , value, projection='Robin', vminv=0, vmaxv=0.5,
                c_colour='YlOrRd', title_lab = '',
                title_str=None,psize=0.5,colorbar_labsize=25,dpi=100):

    # Marie Doutriaux-Boucher - 11 Jan 2022, based on PMAp and AOD stuff
#===========================================

    proj = proj_dic[projection]  # decide the proj see dict before the routine

    fig = plt.figure(figsize=(10,5))
    ax = fig.add_subplot(1, 1, 1, projection=proj)
    ax.coastlines(linewidth=3,color='0.05')
    ax.gridlines(linestyle='--', color='black')

    ax.set_global()

    if title_str:
        plt.title(title_str)

    myticks = np.arange(vminv,vmaxv+0.0001,(vmaxv-vminv)/10)

    mymap = c_colour

    # do the plot
    
    im = ax.scatter(lon[:], lat[:], c=value[:],cmap=mymap,transform=ccrs.PlateCarree(),
                    vmin=vminv,vmax=vmaxv,s=psize,alpha=1)

    # makes the colorbar of the same size as the map
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="3%", pad=0.1, axes_class=maxes.Axes)

    # plot the color_bar
    cbar = fig.colorbar(im,ax=ax,ticks=myticks,label=title_lab,cax = cax)
    

    cbar.set_label(label=title_lab,size=colorbar_labsize,weight='normal') # change size of the colorbar title
                                                   # see https://matplotlib.org/stable/api/text_api.html for font weight
    im.figure.axes[1].tick_params(axis="y", labelsize=colorbar_labsize)  #change the size of colorbar ticks label

    # save the plot
    plt.savefig(plot_file, dpi=dpi,bbox_inches='tight')
    plt.show()
    plt.close()


# ### Read a priori variance-covariance matrix, provided in the section 6.1 of the CO PUIM

Sa_CO_init=pd.read_csv('Sa_CO.csv',header=None,index_col=False)
Sa_CO=Sa_CO_init.values
del(Sa_CO_init)


print(Sa_CO)


# ### Provide one single L2 CO CDR netcdf file to read

# choose the file

year  = '2022'
month = '02'
day   = '22'
sat   = 'M01'

#path_to_co_cdr_file = '/DSNNAS/Repro/iasi/IASI_eum_cdr_r0100/data/'+sat+'/IASCOx2x0100/iasi_co/netcdf/'+year+'/'+month+'/'+day+'/'
path_to_co_cdr_file = './'
orbit_co            = 'W_XX-EUMETSAT-Darmstadt,HYPERSPECT+SOUNDING,METOPB+CO+IASI_C_EUMP_20220222121200Z_20220222130559Z_eps_r_l2_0100.nc'
netcdf_inputfile    = path_to_co_cdr_file+orbit_co

print(netcdf_inputfile)


# ## reading with xarray

xrfile = xr.open_dataset(netcdf_inputfile)


variables = ['forli_layer_heights_co','lat','lon','co_npca','co_qflag','co_cp_co_a','co_cp_air','co_x_co','co_nfitlayers','co_npca','co_h_eigenvectors','co_h_eigenvalues']


xrfile[variables]


along_track  = xrfile.sizes['along_track']
across_track = xrfile.sizes['across_track']
nl_co        = xrfile.sizes['nl_co']
co_levels    = xrfile['forli_layer_heights_co'].values # CO levels in m


print(co_levels)

print(xrfile['co_nfitlayers'].shape)
print(xrfile['co_h_eigenvalues'].shape)
print(xrfile['co_h_eigenvectors'].shape)
print(xrfile['co_npca'].shape)
print(xrfile['co_cp_co_a'].shape)
along_track=xrfile.sizes['along_track']
across_track=xrfile.sizes['across_track']
print(along_track)
print(nl_co)


# ### do some tests to keep only valid profiles

# NEEDED to check later valid profiles  - test 11
#----------------------------------------------
tmp = xrfile['co_h_eigenvalues']
npca_values = xrfile['co_npca'].values

# Replace NaN values with 0 and convert to integer
npca_values[np.isnan(npca_values)] = 0
npca_values = npca_values.astype(int)

for i in range(along_track):
    for j in range(across_track):
        ind = npca_values[i, j]
        tmp[i, j, ind:] = 0


test1 = xrfile['co_nfitlayers']>-1            
test2 = xrfile['co_x_co'].std(axis=2) > 0               # check that co_x_co is not constant (e.g. if standard dev >0)
test3 = (xrfile['co_x_co']==0).sum(axis=2) == 0         # check that co_x_co profile is not full of 0 (sum all level with value 0)
test4 = (xrfile['co_x_co']>9.96e36).sum(axis=2) == 0    # check that co_x_co profile do not contain value > 9.96e3
test5 = (xrfile['co_x_co']<1e-5).sum(axis=2) == 0       # check that profile do not contain value <1E-5
test6 = ((xrfile['co_x_co']>650000) & 
         (xrfile['co_x_co']<660000)).sum(axis=2) == 0   # check that co_x_co profile has no values >650000 and <660000
test7 = (xrfile['co_cp_co_a']==0).sum(axis=2) == 0      # check that co_xp_co_a profile is not full of 0
test8 = xrfile['co_x_co'].isnull().sum(axis=2) == 0     # check that co_c_co has no nan
test9 = (xrfile['lat']<90) & (xrfile['lat']>-90)        # check that lat has valid values
test10 = xrfile['co_qflag']>-1                          # check the flag > -1
test11 = xrfile['co_h_eigenvalues'].sum(axis=2) == xrfile['co_npca']  # check that co_npca (num of vector describing the matrix)
                                                                      # is equal to the num of eigen value !!! need the prior 
                                                                      # rearrangment
keep = xr.where( test1 & test2 & test3 & test4 & test5 & test6 & test7 & test8 & test9 & test10 & test11, True, False)
print(keep.sum())


print(xr)


# ### keep some variables for the 'valid' profiles

keep_indices = ~xrfile['lat'].where(keep).isnull()

lat     = xrfile['lat'].values[keep_indices]
lon     = xrfile['lon'].values[keep_indices]
scalfac = xrfile['co_x_co'].values[keep_indices]
prior   = xrfile['co_cp_co_a'].values[keep_indices]
air     = xrfile['co_cp_air'].values[keep_indices]
fit     = xrfile['co_cp_co_a'].values[keep_indices]*xrfile['co_x_co'].values[keep_indices]
coltot  = np.nansum(fit,axis=1)
nfitlay = xrfile['co_nfitlayers'].values[keep_indices]
eve     = xrfile['co_h_eigenvectors'].values[keep_indices]
eva     = xrfile['co_h_eigenvalues'].values[keep_indices]
npca    = xrfile['co_npca'].values[keep_indices]
surfp   = xrfile['surface_pressure'].values[keep_indices]
gqf     = xrfile['co_qflag'].values[keep_indices]


print(gqf.shape)


print('lat shape:',lat.shape)
print('lon shape:',lat.shape)
print('fit shape',fit.shape)
print('air.shape',air.shape)
print('scalfac.shape',scalfac.shape)
print('coltot shape',coltot.shape)
print('eve.shape',eve.shape)
print('scalfac shape:',scalfac.shape)
print('co_nfitlayers.shape ',nfitlay.shape)
print('npca.shape',npca.shape)
print('prior.shape',prior.shape)
print('eva.shape',eva.shape)
print('surfp.shape',surfp.shape)


# # reconstruct matrices + error estimation - 

ERROR_PROF   = np.tile(np.nan,(len(lat),nl_co))
ERROR_COLTOT = np.tile(np.nan,(len(lat)))

AVK          = np.tile(np.nan,(len(lat),nl_co,nl_co))
dofs         = np.tile(np.nan,(len(lat)))


for i in range(lat.size):
    
    # matrices reconstruction + errors estimation
    tmp_eve   = eve.take(i,axis=0)
    tmp_eve2  = tmp_eve.take(range(0,int(nfitlay[i])*int(npca[i])))
    transp_nu = np.reshape(tmp_eve2,(int(npca[i]),int(nfitlay[i])))
    del(tmp_eve)
    del(tmp_eve2)

    nu         = np.transpose(transp_nu)
    tmp_eva    = eva.take(i,axis=0)
    tmp_eva2   = tmp_eva.take(range(0,int(npca[i])))

    dia_lambda = np.diag(tmp_eva2)
    del(tmp_eva)
    del(tmp_eva2)

    H1 = nu.dot(dia_lambda)
    H  = H1.dot(transp_nu)
    del(H1)

    Sa = Sa_CO.take(range(19-int(nfitlay[i]),19),axis=1)
    Sa = Sa.take(range(19-int(nfitlay[i]),19),axis=0)

    inv_Sa = np.linalg.inv(Sa)
    S      = np.linalg.inv(H+inv_Sa)

    ERROR_PROF1 = np.divide(np.sqrt(np.diag(S)),scalfac[i,range(19-int(nfitlay[i]),19)])
    ERROR_PROF[i,range(19-int(nfitlay[i]),19)] = ERROR_PROF1

    S_CP1 = np.diag(prior[i,range(19-int(nfitlay[i]),19)]).dot(S)
    S_CP  = S_CP1.dot(np.diag(prior[i,range(19-int(nfitlay[i]),19)]))
    del(S_CP1)

    ERROR_COLTOT[i]=np.sqrt(np.sum(np.sum(S_CP)))/np.sum(fit[i,range(19-int(nfitlay[i]),19)])
    del(S_CP)

    AVK_SF = S.dot(H)
    del(H)
    del(S)

    AVK_CP1 = np.diag(prior[i,range(19-int(nfitlay[i]),19)]).dot(AVK_SF)
    AVK_CP  = AVK_CP1.dot(np.linalg.inv(np.diag(prior[i,range(19-int(nfitlay[i]),19)])))
    del(AVK_CP1)

    dofs[i] = np.sum(np.diag(AVK_CP))

    AVK[i,19-int(nfitlay[i,]):,19-int(nfitlay[i]):] = AVK_CP

    del(AVK_CP)


# # plot maps of tot column CO

dirout = '/DSNNAS/Repro_Temp/users/marieb/tmp/'

mini  =  1.E17 #min#220 #180
maxi  =  4.E18 #xmax#320 #310

fileout     = dirout+'test_CO_valid_marie_'+sat+'_'+day+'_'+month+'_'+year+'_1_orb.png'
colorp = 'jet'
proj = 'Robin'
title_str=''
title_str= sat+' '+day+' '+month+' '+year

create_map(fileout, lat, lon,coltot, projection = proj,
            vminv = mini,vmaxv=maxi, c_colour=colorp, title_lab='CO totcol [molecules/cm$^2$]', title_str=title_str,
            psize=5,colorbar_labsize=15)


# # plot one of the good profile

obs = 1    #choose the profile
print(fit[1,:])


# ##  in molecule/cm2/layer of 1 km (fit)

# plot one profile at one location
print(co_levels)
fig, ax1 = plt.subplots(figsize=(7,10))
fontsize=14

xmin = 1.E15
xmax = 2.5E17

ax1.set_ylabel('Altitude (km)',fontsize=22)
#ax1.set_xlabel('')
ax1.set_xlim(xmin,xmax)
ax1.set_ylim(0,20)
plt.xticks(fontsize=fontsize)
ax1.set_yticks([0,2,4,6,8,10,12,14,16,18])
ax1.set_yticklabels([0,2,4,6,8,10,12,14,16,18],fontsize=fontsize)
ax1.set_xlabel('CO (molecules/cm$^2$/layer of 1 km)',fontsize = fontsize+5)

ax1.plot(fit[obs,:-1],(co_levels[:-1]+500)/1000.,color='black',marker="o")
               
for i in range(nl_co):
    ax1.plot([xmin,xmax],[co_levels[i]/1000.,co_levels[i]/1000.],color='grey',linewidth=0.2)

plt.title('lat:'+str(round(lat[obs],4))+', lon:'+str(round(lon[obs],4)))

plt.savefig('Test_CO_prof_19lev.png',dpi=100,bbox_inches='tight')


# ## in ppb : fit/air*1E9

# plot one profile at one location
print('co levels in m:',co_levels)
coltot_ppb = np.sum(fit[obs,:-1])/np.sum(air[obs,:-1])*1e9 # convert in ppb the column
fig, ax1 = plt.subplots(figsize=(7,10))
fontsize=14

xmin = 20
xmax = 60


ax1.set_ylabel('Altitude (km)',fontsize=22)
#ax1.set_xlabel('')
ax1.set_xlim(xmin,xmax)
ax1.set_ylim(0,20)
plt.xticks(fontsize=fontsize)
ax1.set_yticks([0,2,4,6,8,10,12,14,16,18])
ax1.set_yticklabels([0,2,4,6,8,10,12,14,16,18],fontsize=fontsize)
ax1.set_xlabel('CO (ppb)',fontsize = fontsize+5)

ax1.plot(fit[obs,:-1]/air[obs,:-1]*1e9,(co_levels[:-1]+500)/1000.,color='black',marker="o")

plt.title('lat:'+str(round(lat[obs],4))+', lon:'+str(round(lon[obs],4))+', coltot='+str(round(coltot_ppb,3))+' ppb')
               
for i in range(nl_co):
    #i=i+1 
    #print('i=',i,' pp[i]=',pp[i])
    ax1.plot([xmin,xmax],[co_levels[i]/1000.,co_levels[i]/1000.],color='grey',linewidth=0.2)
    

plt.savefig('Test_CO_prof_19lev.png',dpi=100,bbox_inches='tight')


# # some prints for example

obs = 2
print('obs=',obs,'lat = ',lat[obs],' lon=',lon[obs],'surfp = ',surfp[obs])
print('----------')
mmm_print(fit[obs,:-1],'fit = profile in molecule/cm2/layer of 1km: ')
mmm_print(fit[obs,:-1]/air[obs,:-1]*1e9,'fit/air*1e9 = profile in ppb')
print()

print('coltot in ppm from sum of each layers:sum fit/air in ppb',sum(fit[obs,:-1])/sum(air[obs,:-1])*1e9)

print('coltot in molecule/cm2:',coltot[obs])
print()
print('fit',fit[obs,:])

print('fit in ppm per layer of 1km:',fit[obs,:-1]/air[obs,:-1]*1e9)


# # save the valid profiles in a dataframe
if 'df' in globals():
    del df


# create the dataframe
df = pd.DataFrame(lat,columns=['lat'])
# add other columns
df['lon']     = lon
df['surfp']   = surfp[:]
df['gqf']     = gqf[:]
df['nfitlay'] = nfitlay[:]
df['fit']     = list(fit[:,:])
df['fit_ppb'] = list(fit[:,:]/air[:,:]*1e9)
df['ERROR_PROF'] = list(ERROR_PROF[:,:])
df['coltot']  = coltot[:]
df['ERROR_COLTOT'] = ERROR_COLTOT[:]
df['dofs'] = dofs

df.to_csv('Good_profiles.csv', index=False)

# # read the newly creted dataframe
test_df=pd.read_csv('Good_profiles.csv',index_col=False)

print(test_df['fit_ppb'][0])
print(lat[0])
